import unicodedata
from typing import Any, List, Optional

from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from ..base.mapper import SingleBaseMapper
from ..base.types import TransformElementType


class TokenizerMapper(SingleBaseMapper):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        input_field: str,
        output_prefix: Optional[str] = None,
        add_special_tokens: Optional[bool] = True,
        max_length: Optional[int] = None,
        is_split_into_words: Optional[bool] = False,
        return_token_type_ids: Optional[bool] = False,
        return_attention_mask: Optional[bool] = True,
        return_overflowing_tokens: Optional[bool] = False,
        return_special_tokens_mask: Optional[bool] = False,
        return_offsets_mapping: Optional[bool] = False,
        return_length: Optional[bool] = False,
        return_word_ids: Optional[bool] = False,
        return_words: Optional[bool] = False,
        **tk_kwargs: Any,
    ) -> None:
        self.prefix = f"{output_prefix}_" if output_prefix else ""
        self.to_tokenize_filed = input_field
        self.tokenizer = tokenizer

        # arguments to be passed to the tokenizer __call__ function go here
        tk_kwargs = {
            "add_special_tokens": add_special_tokens,
            "max_length": max_length,
            "is_split_into_words": is_split_into_words,
            **(tk_kwargs or {}),
        }

        output_fields = ["input_ids"]

        # various options for the tokenizer affect which fields are returned
        tk_kwargs["return_attention_mask"] = return_attention_mask
        if return_attention_mask:
            output_fields.append("attention_mask")

        tk_kwargs["return_token_type_ids"] = return_token_type_ids
        if return_token_type_ids:
            output_fields.append("token_type_ids")

        tk_kwargs["return_overflowing_tokens"] = return_overflowing_tokens
        if return_overflowing_tokens:
            output_fields.append("overflow_to_sample_mapping")

        tk_kwargs["return_special_tokens_mask"] = return_special_tokens_mask
        if return_special_tokens_mask:
            output_fields.append("special_tokens_mask")

        tk_kwargs["return_offsets_mapping"] = return_offsets_mapping
        if return_offsets_mapping:
            output_fields.append("offset_mapping")

        tk_kwargs["return_length"] = return_length
        if return_length:
            output_fields.append("length")

        self.return_word_ids = self.return_words = False
        if "is_split_into_words" in tk_kwargs and return_word_ids:
            self.return_word_ids = True
            output_fields.append("word_ids")
            if return_words:
                self.return_words = True
                output_fields.append("words")

        self.tokenize_kwargs = tk_kwargs
        super().__init__(
            input_fields=[self.to_tokenize_filed],
            output_fields=list(map(self._prefixify, output_fields)),
        )

    def _prefixify(self, field_or_dict: str) -> str:
        return f"{self.prefix}{field_or_dict}"

    def transform(self, data: TransformElementType) -> TransformElementType:
        batch_encoding = self.tokenizer(
            data[self.to_tokenize_filed], **self.tokenize_kwargs
        )

        # token_to_word mappings are unfortunately not natively returned by
        # HF.tokenizer; so we need to operate separately on the
        # `batch_encoding` object to get this info.
        # if f'{self.prefix}word_ids' in self.output_fields:
        if self.return_word_ids:
            batch_encoding["word_ids"] = [
                # ignoring pylance complaining because word ids should
                # be provided given that we are passing `is_split_into_words`
                batch_encoding[sequence_id].word_ids  # type: ignore
                for sequence_id in range(
                    # ignoring pylance complaining because input_ids should
                    # always be a list of lists when `is_split_into_words`
                    # is set to True.
                    len(batch_encoding["input_ids"])  # type: ignore
                )
            ]
            if self.return_words:
                batch_encoding["words"] = [
                    [
                        None
                        if word_id is None
                        else data[self.input_fields[0]][word_id]
                        for word_id in word_ids
                    ]
                    # ignoring pylance complaining because we just added
                    # word_ids above!
                    for word_ids in batch_encoding["word_ids"]  # type: ignore
                ]

        return {
            self._prefixify(field_name): field_value
            for field_name, field_value in batch_encoding.items()
        }


class ValidUnicodeMapper(SingleBaseMapper):
    """Given input_fields of type List[str], replaces invalid Unicode
    characters with something else"""

    def __init__(
        self,
        input_fields: List[str],
        unicode_categories: List[str],
        replace_token: str,
        # output_prefix: Optional[str] = None,
    ):
        # self.batched = False
        # self.input_fields = input_fields
        # self.prefix = f'{output_prefix}_' if output_prefix else ''
        # self.output_fields = [
        #     f'{self.prefix}{input_field}' for input_field in self.input_fields
        # ]
        super().__init__(input_fields=input_fields, output_fields=input_fields)
        self.unicode_categories = unicode_categories
        self.replace_token = replace_token

    def transform(self, data: TransformElementType) -> TransformElementType:
        def _transform(tokens: List[str]) -> List[str]:
            return [
                self.replace_token
                if all(
                    unicodedata.category(ch) in self.unicode_categories
                    for ch in token
                )
                else token
                for token in tokens
            ]

        return {
            k: v if k not in self.input_fields else _transform(v)
            for k, v in data.items()
        }
        # new_data = {f'{self.prefix}{k}': v if k not in self.input_fields
        #             else _transform(v) for k, v in data.items()}
        # return new_data



from .dataset import BaseDataset
from .mapper import BaseMapper
from .types import (
    Features,
    FeatureType,
    TransformBatchType,
    TransformElementType,
)

__all__ = [
    "BaseDataset",
    "BaseMapper",
    "Features",
    "FeatureType",
    "TransformBatchType",
    "TransformElementType",
]



from typing import Callable, Optional, Protocol, TypeVar

from .types import Features, FeatureType

D = TypeVar("D", bound="BaseDataset")


class BaseDataset(Protocol):
    """A protocol for datasets. In general, a dataset is a collection of
    samples that can be iterated over. Datasets support a map operation that
    can either transform a sample into another sample (1-to-1 mapping) or
    generate multiple samples from a single sample (1-to-many mapping).

    Dataset APIs are modeled after HuggingFace ArrowDataset APIs."""

    @property
    def features(self) -> Features:
        ...

    def map(
        self: D,
        function: Optional[Callable] = None,
        *,
        batched: bool = False,
        batch_size: Optional[int] = 1000
    ) -> D:
        ...

    def cast_column(self: D, column: str, feature: FeatureType, *_) -> D:
        ...



from typing import Any, Callable, List, Optional, TypeVar

from ..base.dataset import BaseDataset
from ..base.mapper import SingleBaseMapper
from ..base.types import TransformElementType

D = TypeVar("D", bound="BaseDataset")


class ChangeFieldsMapper(SingleBaseMapper):
    def __init__(
        self,
        keep_fields: Optional[List[str]] = None,
        drop_fields: Optional[List[str]] = None,
    ):
        """Mapper that removes some of the fields in a dataset.
        Either `keep_fields` or `drop_fields` must be specified, but not both.

        Args:
            keep_fields (List[str]): Fields to keep, all other fields
                are dropped. Defaults to [].
            drop_fields (List[str]): Fields to drop, all other fields
                are kept. Defaults to [].
        """

        # xor between keep_fields and remove_fields
        if (keep_fields is not None and drop_fields is not None) or (
            keep_fields is None and drop_fields is None
        ):
            raise ValueError("Must specify `keep_fields` or `drop_fields`")

        super().__init__(input_fields=drop_fields, output_fields=keep_fields)

    def map(self, dataset: D, **map_kwargs: Any) -> D:
        map_kwargs = {
            "remove_columns": list(dataset.features.keys()),
            **map_kwargs,
        }
        return super().map(dataset, **map_kwargs)

    def transform(self, data: TransformElementType) -> TransformElementType:
        if self.input_fields:
            new_data = {
                k: v for k, v in data.items() if k not in self.input_fields
            }

        elif self.output_fields:
            new_data = {k: data[k] for k in self.output_fields}

        else:
            raise ValueError("Must specify `keep_fields` or `drop_fields`")

        return new_data


class MakeFieldMapper(SingleBaseMapper):
    def __init__(
        self,
        field_name: str,
        value: Optional[Any] = None,
        shape_like: Optional[str] = None,
        value_fn: Optional[Callable[[TransformElementType], Any]] = None,
    ):
        """Mapper that adds a new field to a dataset.
        Either `value` or `value_fn` must be specified, but not both.
        `value` is a constant value to assign to the new field, while
        `value_fn` is a function that takes the full sample as input and
        returns a value to assign to the new field.

        Args:
            field_name (str): Name of the new field.
            value (Optional[Any], optional): Value to assign to the new field.
            shape_like (Optional[str], optional): If a fixed value is provided,
                this existing field that will be used to determine the shape of
                the new field. Defaults to None.
            value_fn (Optional[Callable[[TransformElementType], Any]], optional):
                Function to call to assign a value to the new field.
                Defaults to None.
        """
        super().__init__(output_fields=[field_name])

        if value_fn is None:
            if value is None:
                raise ValueError("Must specify `value` or `value_fn`")

            def _value_fn(data: TransformElementType) -> Any:
                if shape_like is not None:
                    return [value for _ in data[shape_like]]
                else:
                    return value

            # doing the two assignment separately to avoid linter
            # error about redefinition of `value_fn` via def
            value_fn = _value_fn

        self.value_fn = value_fn

    def transform(self, data: TransformElementType) -> TransformElementType:
        data[self.output_fields[0]] = self.value_fn(data)
        return data



from typing import Any, Callable, Optional, Sequence

from ..base.dataset import BaseDataset
from ..base.mapper import BaseMapper
from ..base.types import (
    Features,
    FeatureType,
    TransformBatchType,
    TransformElementType,
)
from ..mappers import fields, multiseq, shape, tokenize
from ..mappers.contrib import sse

__all__ = [
    "Dataset",
    "TokensSequencesPaddingMapper",
    "AttentionMaskSequencePaddingMapper",
    "TokenTypeIdsSequencePaddingMapper",
    "MakeAttentionMaskMapper",
    "LabelsMaskerMapper",
    "MultiSequenceStriderMapper",
    "SingleValueToSequenceMapper",
    "SequencesConcatenateMapper",
    "FlattenMapper",
    "BinarizerMapper",
    "TokenizerMapper",
    "OneVsOtherAnnotatorMapper",
    "ChangeFieldsMapper",
    "MakeFieldMapper",
    "ValidUnicodeMapper",
]


class Dataset(list, BaseDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def features(self) -> Features:
        if len(self):
            return {k: type(v) for k, v in self[0].items()}
        else:
            raise ValueError("Dataset is empty; cant get features")

    def keys(self) -> Sequence[str]:
        return [k for k in self.features.keys()]

    def map(
        self: "Dataset",
        function: Optional[Callable] = None,
        batched: bool = False,
        *args,
        **kwargs,
    ) -> "Dataset":
        if function is None:
            return self

        processed = Dataset()
        for sample in self:
            if batched:
                processed.extend(function([sample]))
            else:
                processed.append(function(sample))

        return processed

    def cast_column(
        self: "Dataset", column: str, feature: FeatureType, *_
    ) -> "Dataset":
        def cast_fn(sample: TransformElementType) -> TransformElementType:
            return {
                k: (feature(v) if k == column else v)
                for k, v in sample.items()
            }

        return self.map(cast_fn)


class _SimpleInterfaceMixInMapper(BaseMapper):
    def map(  # type: ignore
        self: "_SimpleInterfaceMixInMapper",
        dataset: Dataset,
        **map_kwargs: Any,
    ) -> Dataset:
        # this function is re-implemented to provide nice
        # type annotations.

        if not isinstance(dataset, Dataset):
            raise ValueError(
                f"Dataset must be of type `{Dataset}`"
                f" but I got `{type(dataset)}` instead."
            )

        # This ensures that the super for the next in line in the mro,
        # which should be whatever actual implementation of the mapper
        # this mixin is mixed in with, will be called.
        return super().map(dataset, **map_kwargs)

    @property
    def batched(self: "_SimpleInterfaceMixInMapper") -> bool:
        # This ensures that the super for the next in line in the mro,
        # which should be whatever actual implementation of the mapper
        # this mixin is mixed in with, will be called.
        return super().batched

    def _batch_transform(  # type: ignore
        self,  # type: ignore
        data: TransformBatchType,  # type: ignore
    ) -> TransformBatchType:  # type: ignore
        # Simple mappers are called with slightly different data from
        # the base class, so we need to override this method.

        # Note that we need to disable the type check to avoid a mess
        # with mypy. It's a bit of a hack, because the type returned here is
        # not what mypy expects, but it works as long as Dataset is an
        # instance of a list.

        yield from super().transform(data)


class TokensSequencesPaddingMapper(
    _SimpleInterfaceMixInMapper, multiseq.TokensSequencesPaddingMapper
):
    ...


class AttentionMaskSequencePaddingMapper(
    _SimpleInterfaceMixInMapper, multiseq.AttentionMaskSequencePaddingMapper
):
    ...


class TokenTypeIdsSequencePaddingMapper(
    _SimpleInterfaceMixInMapper, multiseq.TokenTypeIdsSequencePaddingMapper
):
    ...


class MakeAttentionMaskMapper(
    _SimpleInterfaceMixInMapper, multiseq.MakeAttentionMaskMapper
):
    ...


class LabelsMaskerMapper(
    _SimpleInterfaceMixInMapper, multiseq.LabelsMaskerMapper
):
    ...


class MultiSequenceStriderMapper(
    _SimpleInterfaceMixInMapper, multiseq.MultiSequenceStriderMapper
):
    ...


class SingleValueToSequenceMapper(
    _SimpleInterfaceMixInMapper, multiseq.SingleValueToSequenceMapper
):
    ...


class SequencesConcatenateMapper(
    _SimpleInterfaceMixInMapper, multiseq.SequencesConcatenateMapper
):
    ...


class FlattenMapper(_SimpleInterfaceMixInMapper, shape.FlattenMapper):
    ...


class BinarizerMapper(_SimpleInterfaceMixInMapper, shape.BinarizerMapper):
    ...


class TokenizerMapper(_SimpleInterfaceMixInMapper, tokenize.TokenizerMapper):
    ...


class OneVsOtherAnnotatorMapper(
    _SimpleInterfaceMixInMapper, sse.OneVsOtherAnnotatorMapper
):
    ...


# We ignore type error bc they are due to the fact that
# the `map` method here maps to a BaseDataset vs the Dataset
# type expected by _SimpleInterfaceMixInMapper.
class ChangeFieldsMapper(  # type: ignore
    _SimpleInterfaceMixInMapper, fields.ChangeFieldsMapper
):
    ...


class MakeFieldMapper(_SimpleInterfaceMixInMapper, fields.MakeFieldMapper):
    ...


class ValidUnicodeMapper(
    _SimpleInterfaceMixInMapper, tokenize.ValidUnicodeMapper
):
    ...



from abc import ABCMeta, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from .dataset import BaseDataset
from .types import (
    Features,
    FeatureType,
    TransformBatchType,
    TransformElementType,
)

if TYPE_CHECKING:
    from .pipeline import Pipeline

__all__ = ["SingleBaseMapper", "BatchedBaseMapper"]


DatasetType = TypeVar("DatasetType", bound="BaseDataset")


class BaseMapper(metaclass=ABCMeta):
    """An abstract implementation of a Mapper. Do not subclass directly;
    instead subclass either SimpleBaseMapper (if your mapper operates on
    a single element) or BatchedBaseMapper (if your mapper operates on a
    batch).

    A mapper is an object that transforms samples in a dataset. All mappers
    do so by implementing a method called `transform`.

    Optionally, mappers can also implement the cast_columns method,
    which returns a dictionary of new types for some (or all) of the
    input/output columns.
    """

    __slots__: List[str] = ["input_fields", "output_fields"]
    input_fields: List[str]
    output_fields: List[str]

    def __init__(
        self: "BaseMapper",
        input_fields: Optional[List[str]] = None,
        output_fields: Optional[List[str]] = None,
    ) -> None:
        """Create a new Mapper.

        Args:
            input_fields (Optional[List[str]], optional): The fields expected
                by this mapper. If None is provided, the mapper will not
                check for the presence of any input fields. Defaults to None.
            output_fields (Optional[List[str]], optional): The fields produced
                by this mapper after transformation. If None is provided, the
                mapper will not validate the output of transform. Defaults to
                None.
        """
        self.input_fields = input_fields or []
        self.output_fields = output_fields or []

    @property
    @abstractmethod
    def batched(self: "BaseMapper") -> bool:
        raise NotImplementedError("batched property must be implemented")

    def _verify_output(self: "BaseMapper", data: Dict[str, Any]):
        for field in self.output_fields:
            if field not in data:
                raise ValueError(f"Field {field} not found in returned data")

    def __lshift__(
        self: "BaseMapper", other: Union["BaseMapper", "Pipeline"]
    ) -> "Pipeline":
        """Create a new Pipeline by combining this mapper with another."""

        # import actual implementation here to avoid circular imports
        from .pipeline import Pipeline

        return Pipeline(self) << other

    def __rshift__(self: "BaseMapper", other: "BaseMapper") -> "Pipeline":
        """Create a new Pipeline by combining this mapper with another."""
        return other << self

    def map(
        self: "BaseMapper", dataset: DatasetType, **map_kwargs: Any
    ) -> DatasetType:
        """Transform a dataset by applying this mapper's transform method.

        Args:
            dataset (DatasetType): The dataset to transform.
            **map_kwargs: Additional keyword arguments to pass to the
                transform method. By default, this is empty.

        Returns:
            dataset (DatasetType): The transformed dataset.
        """

        for field in self.input_fields:
            if field not in dataset.features:
                msg = f"Field {field} not found in input dataset {dataset}"
                raise ValueError(msg)

        map_kwargs = {"batched": self.batched, **map_kwargs}

        if self.batched:
            dataset = dataset.map(function=self._batch_transform, **map_kwargs)
        else:
            dataset = dataset.map(function=self.transform, **map_kwargs)

        # We optionally cast some (or all!) of the output columns
        # to a new format. By default, no casting occurs.
        to_cast_features = self.cast_columns(dataset.features)
        for feat_name, feat_type in to_cast_features.items():
            dataset = dataset.cast_column(feat_name, feat_type)

        for field in self.output_fields:
            if field not in dataset.features:
                raise ValueError(f"Field {field} not found in returned data")
        return dataset

    @abstractmethod
    def _batch_transform(
        self: "BaseMapper", data: TransformBatchType
    ) -> TransformBatchType:
        """Internal method for transforming a batch of data; this method
        is called by the map method when the mapper is batched.

        Any actual mapper implementation should NOT override this method;
        instead, any  mapper that operates on a batch should override the
        transform method, where the transform method receives a iterable
        of sample dictionaries as input.

        Args:
            data (TransformBatchType): The batch of data to transform.

        Returns:
            TransformBatchType: A batch of transformed data.
        """
        raise NotImplementedError(
            "Mapper subclass must implement _batch_transform"
        )

    @abstractmethod
    def transform(self: "BaseMapper", data: Any) -> Any:
        """Transform a single sample of a dataset. This method should be
        overridden by actual mapper implementations.

        Args:
            self (BaseMapper): _description_
            data (Any): _description_

        Raises:
            NotImplementedError: _description_

        Returns:
            Any: _description_
        """
        raise NotImplementedError("Mapper subclass must implement transform")

    def cast_columns(
        self: "BaseMapper", features: Features
    ) -> Dict[str, FeatureType]:
        return {}


class SingleBaseMapper(BaseMapper, metaclass=ABCMeta):
    """An abstract implementation of a Mapper that operates on a single
    element. All mappers that operate on a single element should subclass
    this class.

    Actual mapper implementations should override the transform method.
    The transform method should accept a single sample dictionary as input
    and return a single sample dictionary as output.
    """

    @property
    def batched(self) -> bool:
        return False

    def _batch_transform(
        self: "BaseMapper", data: TransformBatchType
    ) -> TransformBatchType:
        raise ValueError("SingleBaseMapper does not support batching")

    @abstractmethod
    def transform(
        self: "BaseMapper", data: TransformElementType
    ) -> TransformElementType:
        """Transform a single sample of a dataset. This method should be
        overridden by actual mapper implementations.

        Args:
            data (TransformElementType): The sample to transform. This is
                a single sample dictionary with string keys and values of
                any type.

        Returns:
            TransformElementType: The transformed sample. This is a single
                sample dictionary with string keys and values of any type.
                The keys can be different from the input keys.
        """
        raise NotImplementedError("Mapper subclass must implement transform")


class BatchedBaseMapper(BaseMapper, metaclass=ABCMeta):
    """An abstract implementation of a Mapper that operates on a batch of
    elements. All mappers that operate on a batch should subclass this
    class.

    Actual mapper implementations should override the transform method.
    The transform method should accept a iterator of dictionaries as input,
    and return a iterator of dictionaries as output. The number of samples
    returned may be different from the number of samples in the input.
    """

    @property
    def batched(self) -> bool:
        return True

    def _batch_transform(
        self: "BatchedBaseMapper", data: TransformBatchType
    ) -> TransformBatchType:
        """Internal method for transforming a batch of data; this method
        is called by the map method when the mapper is batched.

        Any actual mapper implementation should NOT override this method;
        instead, any  mapper that operates on a batch should override the
        transform method instead."""

        keys = [k for k in data.keys()]

        def _index_fn(t: Tuple[str, Any]) -> int:
            k, _ = t
            return keys.index(k)

        to_transform_iterable = (
            dict(zip(keys, sample))
            for sample in zip(
                *(v for _, v in sorted(data.items(), key=_index_fn))
            )
        )
        transformed_batch: Dict[str, List[Any]] = {}
        for transformed_sample in self.transform(to_transform_iterable):
            for k, v in transformed_sample.items():
                transformed_batch.setdefault(k, []).append(v)

        return transformed_batch

    @abstractmethod
    def transform(
        self: "BatchedBaseMapper", data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:
        """Transform a batch of data. This method should be overridden by
        actual mapper implementations.

        Args:
            data (Iterable[TransformElementType]): The batch of data to
                transform; it is an iterable of dictionaries, where each
                dictionary is a sample (i.e. a Dict[str, Any]).

        Returns:
            Iterable[TransformElementType]: Iterable of transformed data.
                Each element in the iterable is a dictionary with str keys
                and values of any type. The number of elements in the
                iterable may be different from the number of samples in the
                input.
        """
        raise NotImplementedError("Mapper subclass must implement transform")



from itertools import chain
from typing import Dict

from ..base.mapper import SingleBaseMapper
from ..base.types import Features, FeatureType, TransformElementType


class FlattenMapper(SingleBaseMapper):
    def __init__(self, field: str) -> None:
        super().__init__(input_fields=[field], output_fields=[field])

    def transform(self, data: TransformElementType) -> TransformElementType:
        field_name, *_ = self.input_fields

        flattened_field = data[field_name]

        if len(flattened_field) > 0:
            while isinstance(flattened_field[0], list):
                flattened_field = list(chain.from_iterable(flattened_field))

        return {field_name: flattened_field}


class BinarizerMapper(SingleBaseMapper):
    __value_type__: type = str
    __sequence_type__: type = list

    def __init__(self, field: str, threshold: float) -> None:
        super().__init__(input_fields=[field], output_fields=[field])
        self.threshold = threshold

    def cast_columns(self, features: Features) -> Dict[str, FeatureType]:
        field_name, *_ = self.input_fields
        if isinstance(features[field_name], self.__sequence_type__):
            new_field = self.__sequence_type__(self.__value_type__("int64"))
        else:
            new_field = self.__value_type__("int64")

        return {field_name: new_field}

    def transform(self, data: TransformElementType) -> TransformElementType:
        field_name, *_ = self.input_fields

        if isinstance(data[field_name], list):
            return {
                field_name: [
                    1 if v > self.threshold else 0 for v in data[field_name]
                ]
            }
        else:
            return {field_name: 1 if data[field_name] > self.threshold else 0}



from typing import Any, Dict, TypeVar

from smashed.base.dataset import BaseDataset

from ..base import BaseMapper
from ..mappers import fields, multiseq, shape, tokenize
from ..mappers.contrib import sse
from ..utils import requires

requires("datasets")

# we add the noqa bit because datasets is not part of the core requirement.
# Because we want to fail gracefully, `requires` above checks if `datasets`
# is installed, raising an helpful error message if it is not.
# However, that upsets the linter, which thinks that we should import in order
# instead. Therefore, we slap a bunch of `noqa`s in here to suppress it :)
from datasets import features  # noqa: E402
from datasets.arrow_dataset import Dataset  # noqa: E402
from datasets.features.features import Features, FeatureType  # noqa: E402
from datasets.iterable_dataset import IterableDataset  # noqa: E402


class HfDatasetProtocol(Dataset, BaseDataset):
    ...


class HfIterableDatasetProtocol(IterableDataset, BaseDataset):
    ...


HfDatasetType = TypeVar(
    "HfDatasetType", HfDatasetProtocol, HfIterableDatasetProtocol
)


__all__ = [
    "TokensSequencesPaddingMapper",
    "AttentionMaskSequencePaddingMapper",
    "TokenTypeIdsSequencePaddingMapper",
    "MakeAttentionMaskMapper",
    "LabelsMaskerMapper",
    "MultiSequenceStriderMapper",
    "SingleValueToSequenceMapper",
    "SequencesConcatenateMapper",
    "FlattenMapper",
    "BinarizerMapper",
    "TokenizerMapper",
    "OneVsOtherAnnotatorMapper",
    "ChangeFieldsMapper",
    "ValidUnicodeMapper",
]


class _HuggingFaceInterfaceMixInMapper(BaseMapper):
    def map(  # type: ignore
        self, dataset: HfDatasetType, **map_kwargs: Any
    ) -> HfDatasetType:
        # this function is re-implemented to provide nice type annotations.
        # therefore, we remove the mypy type warnings with the ignore above.

        return super().map(dataset, **map_kwargs)

    def cast_columns(self, features: Features) -> Dict[str, FeatureType]:
        return super().cast_columns(features)


class TokensSequencesPaddingMapper(
    _HuggingFaceInterfaceMixInMapper, multiseq.TokensSequencesPaddingMapper
):
    ...


class AttentionMaskSequencePaddingMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.AttentionMaskSequencePaddingMapper,
):
    ...


class TokenTypeIdsSequencePaddingMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.TokenTypeIdsSequencePaddingMapper,
):
    ...


class MakeAttentionMaskMapper(
    _HuggingFaceInterfaceMixInMapper, multiseq.MakeAttentionMaskMapper
):
    ...


class LabelsMaskerMapper(
    _HuggingFaceInterfaceMixInMapper, multiseq.LabelsMaskerMapper
):
    ...


class MultiSequenceStriderMapper(
    _HuggingFaceInterfaceMixInMapper, multiseq.MultiSequenceStriderMapper
):
    ...


class SingleValueToSequenceMapper(
    _HuggingFaceInterfaceMixInMapper, multiseq.SingleValueToSequenceMapper
):
    ...


class SequencesConcatenateMapper(
    _HuggingFaceInterfaceMixInMapper, multiseq.SequencesConcatenateMapper
):
    ...


class FlattenMapper(_HuggingFaceInterfaceMixInMapper, shape.FlattenMapper):
    ...


class BinarizerMapper(_HuggingFaceInterfaceMixInMapper, shape.BinarizerMapper):
    __value_type__: type = features.Value
    __sequence_type__: type = features.Sequence


class TokenizerMapper(
    _HuggingFaceInterfaceMixInMapper, tokenize.TokenizerMapper
):
    ...


class OneVsOtherAnnotatorMapper(
    _HuggingFaceInterfaceMixInMapper, sse.OneVsOtherAnnotatorMapper
):
    ...


# We ignore type error bc they are due to the fact that the `map` method here
# maps to a BaseDataset vs the Dataset/IterableDataset type expected
# by _HuggingFaceInterfaceMixInMapper.
class ChangeFieldsMapper(  # type: ignore
    _HuggingFaceInterfaceMixInMapper, fields.ChangeFieldsMapper
):
    ...


class ValidUnicodeMapper(
    _HuggingFaceInterfaceMixInMapper, tokenize.ValidUnicodeMapper
):
    ...
